from collections import defaultdict
from itertools import islice
from math import log
from typing import Dict, Iterator, List, Optional

__all__ = [
    'primes_iter', 'primes', 'primes_up_to_iter', 'primes_up_to',
    'mersenne_primes_iter', 'mersenne_primes', 'mersenne_primes_up_to_iter',
    'mersenne_primes_up_to',
]


def yield_values_up_to(n: int, _iter: Iterator[int]):
    num = next(_iter)
    while num <= n:
        yield n
        num = next(_iter)


def primes_iter(n: Optional[int] = None) -> Iterator[int]:
    def infinite_primes() -> Iterator[int]:
        D: Dict[int, List[int]] = defaultdict(list)
        q = 2
        while True:
            if q not in D:
                yield q
                D[q * q] = [q]
            else:
                for p in D[q]:
                    D[p + q].append(p)
                del D[q]
            q += 1

    if n is None:
        return infinite_primes()
    return islice(infinite_primes(), n)


def primes(n: Optional[int] = None) -> List[int]:
    return list(primes_iter(n))


def primes_up_to_iter(n: int) -> Iterator[int]:
    return yield_values_up_to(n, primes_iter())


def primes_up_to(n: int) -> List[int]:
    return list(primes_up_to_iter(n))


def mersenne_primes_iter(n: Optional[int] = None) -> Iterator[int]:
    def infinite_mersenne_primes() -> Iterator[int]:
        D: Dict[int, List[int]] = defaultdict(list)
        q = 2
        while True:
            if q not in D:
                if log(q + 1, 2).is_integer():
                    yield q
                D[q * q] = [q]
            else:
                for p in D[q]:
                    D[p + q].append(p)
                del D[q]
            q += 1

    if n is None:
        return infinite_mersenne_primes()
    return islice(infinite_mersenne_primes(), n)


def mersenne_primes(n: Optional[int] = None) -> List[int]:
    return list(mersenne_primes_iter(n))


def mersenne_primes_up_to_iter(n: int) -> Iterator[int]:
    return yield_values_up_to(n, mersenne_primes_iter())


def mersenne_primes_up_to(n: int) -> List[int]:
    return list(mersenne_primes_iter(n))



import math
from functools import reduce
from math import log
from typing import Iterator, List, Tuple

from . import number_types

__all__ = [
    'are_coprime', 'coprimes', 'xgcd', 'xlcm', 'factors_iter', 'factors', 'gcd',
    'largest_prime_factor', 'lcm', 'phi', 'prime_factors', 'prime_factors_iter',
    'coprimes_iter',
]


def factors_iter(n: int) -> Iterator[int]:
    n = abs(n)
    return (i for i in range(1, n + 1) if n % i == 0)


def factors(n: int) -> List[int]:
    return list(factors_iter(n))


def lcm(a: int, *terms: int) -> int:
    def lcm_binary(a: int, b: int) -> int:
        return a * b // math.gcd(a, b)

    return reduce(lcm_binary, [a] + list(terms))


def gcd(a: int, *terms: int) -> int:
    return reduce(math.gcd, [a] + list(terms))


def xlcm(a: int, *terms: int) -> List[int]:
    _lcm = lcm(a, *terms)
    return [_lcm] + list(map(lambda a: _lcm // a, [a] + list(terms)))


def xgcd(b: int, a: int) -> Tuple[int, int, int]:
    x0, x1, y0, y1 = 1, 0, 0, 1
    while a != 0:
        q, b, a = b // a, a, b % a
        x0, x1 = x1, x0 - q * x1
        y0, y1 = y1, y0 - q * y1
    return b, x0, y0


def prime_factors_iter(n: int) -> Iterator[int]:
    i = 2
    while i * i <= n:
        if n % i != 0:
            i += 1
        else:
            n //= i
            yield i
    if n > 1:
        yield n


def prime_factors(n: int) -> List[int]:
    return list(prime_factors_iter(n))


def largest_prime_factor(n: int) -> int:
    i = 2
    while i * i <= n:
        if n % i:
            i += 1
        else:
            n //= i
    return n


def are_coprime(a: int, b: int, *terms: int) -> bool:
    return reduce(math.gcd, [a] + [b] + list(terms)) == 1


def phi(n: int) -> int:
    """Returns |{x in N : x < n and x coprimes to n}|."""
    if n == 0:
        return 0
    if n == 1:
        return 1
    if number_types.is_prime(n):
        return n - 1
    if log(n, 2).is_integer():
        return n // 2
    return sum(1 for i in range(1, n) if math.gcd(n, i) == 1)


def coprimes_iter(n: int) -> Iterator[int]:
    """Returns {x in N : x < n and x coprime to n}."""
    if number_types.is_prime(n):
        return iter(range(1, n))
    if log(n, 2).is_integer():
        return iter(range(1, n, 2))
    return (i for i in range(1, n) if math.gcd(n, i) == 1)


def coprimes(n: int) -> List[int]:
    return list(coprimes(n))



from typing import Dict, List, Tuple

__all__ = ['benfords_law', 'convert_base', 'rank']


def convert_base(n: int, b: int) -> List[int]:
    """Returns n converted from base 10 to base b."""
    output = []
    while n >= 1:
        n, r = divmod(n, b)
        output.append(r)
    output.reverse()
    return output


def benfords_law(_list: List[float], first_digit=True) -> \
        Dict[int, Tuple[int, float]]:
    """first_digit determines if we are counting the leftmost digit or the
     rightmost digit"""
    count: Dict[int, int] = {}
    start = 1 if first_digit else 0
    for i in range(start, 10):
        count[i] = 0
    index = 0 if first_digit else -1
    for item in _list:
        count[int(str(item)[index])] += 1
    n = len(_list)
    return {key: (value, value / n) for (key, value) in count.items()}


def rank(_list: List[float], dense=False) -> List[int]:
    rank_dict: Dict[float, int] = {}
    if dense:
        i = 1
        for val in sorted(_list):
            if val not in rank_dict:
                rank_dict[val] = i
                i += 1
    else:
        for i, val in enumerate(sorted(_list)):
            if val not in rank_dict:
                rank_dict[val] = i + 1
    return [rank_dict[i] for i in _list]



from typing import Optional, Tuple

from . import number_types, primes

__all__ = ['goldbach', 'goldbach_triple']


def goldbach(n: int) -> Optional[Tuple[int, int]]:
    if n <= 2 or number_types.is_odd(n):
        return None
    _primes = list(primes.primes_up_to(n))
    for p1 in _primes:
        for p2 in reversed(_primes):
            if p1 + p2 == n:
                return p1, p2
            if p1 + p2 < n:
                break
    return None


def goldbach_triple(n: int) -> Optional[Tuple[int, int, int]]:
    if n <= 5:
        return None
    _primes = list(primes.primes_up_to(n))
    for p1 in _primes:
        for p2 in _primes:
            for p3 in _primes:
                if p1 + p2 + p3 == n:
                    return p1, p2, p3
                if p1 + p2 + p3 > n:
                    break
    return None
